# -*- coding: utf-8 -*-
"""
Roseus colormap generator
https://github.com/dofuuz/roseus

Roseus: Perceptually uniform colormaps with full range of lightness
"""

import bisect

import colour
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.colors import ListedColormap


RES = 2 ** 20
POINTS = 256


def gen_colormap(hue_range=(-185, 170), chroma_shape='cos', lightness_range=(2, 99), plots=False):
    """ Generate a perceptually uniform colormap has symmetric and smooth chroma, hue transition.

    Parameters
    ----------
    hue_range : tuple
        Hue range in degrees.
    chroma_shape :
        Shape of chroma transition.
        One of 'sin', 'circle', 'cos'.
    lightness_range : tuple
        Lightness range
    plots :
        Set `True` to print details.
        Set 'verbose' to print additional measures.

    Returns
    -------
        Tuple of (color_rgb, score).
        `color_rgb` is RGB values of the colormap.
        `score` is value represents how much colormap is colorful.
    """
    rx = np.linspace(0, 1, RES)

    if chroma_shape == 'square':
        # https://dsp.stackexchange.com/a/56529
        delta = 0.2
        c = 1 / np.arctan(1/delta) * np.arctan(np.sin(np.pi * rx) / delta)

    elif chroma_shape == 'sin':
        c = np.sin(np.pi * rx)

    elif chroma_shape == 'circle':
        c = np.sqrt(1 - (1 - 2*rx) * (1 - 2*rx))

    elif chroma_shape == 'cos':  # cardioid
        c = (1 - np.cos(2 * np.pi * rx)) / 2

    h = np.linspace(hue_range[0], hue_range[1], RES)
    h_rad = np.radians(h)

    # get ab arc
    a = c * np.cos(h_rad)
    b = c * np.sin(h_rad)

    # differentiate ab arc
    diff_ab = np.hypot(np.diff(a), np.diff(b))
    cumsum_ab = np.cumsum(diff_ab)
    sum_len = cumsum_ab[-1]

    last_cumsum = 0
    idx_sel = []
    idx = 0
    for cdx in range(POINTS):
        length = sum_len * cdx / POINTS
        while True:
            cumsum = cumsum_ab[idx]
            if last_cumsum <= length < cumsum:
                break
            idx += 1

        if length - last_cumsum < cumsum - length:
            idx_sel.append(idx)
        else:
            idx_sel.append(idx+1)

    j = np.linspace(lightness_range[0], lightness_range[1], POINTS)

    def in_gamut(c_mul):
        c_ = c[idx_sel] * c_mul
        h_ = h[idx_sel]
        jch = np.stack([j, c_, h_], axis=-1)

        jab = colour.models.JCh_to_Jab(jch)
        xyz = colour.CAM16UCS_to_XYZ(jab)
        color_rgb = colour.XYZ_to_sRGB(xyz)

        return 0 <= np.min(color_rgb) and np.max(color_rgb) <= 1

    # search c_mul threshold
    c_mul_space = np.arange(45, 20, -0.1)
    cx = bisect.bisect(c_mul_space, False, key=in_gamut)
    if len(c_mul_space) <= cx:
        return None, sum_len * 20 * (lightness_range[1] - lightness_range[0])

    # desaturate into sRGB gamut
    c_mul = c_mul_space[cx]
    c_ = c[idx_sel] * c_mul
    h_ = h[idx_sel]
    jch = np.stack([j, c_, h_], axis=-1)

    jab = colour.models.JCh_to_Jab(jch)
    xyz = colour.CAM16UCS_to_XYZ(jab)
    color_rgb = colour.XYZ_to_sRGB(xyz)

    arc_len = sum_len * c_mul * (lightness_range[1] - lightness_range[0])

    if plots:
        print(f'Max chroma: {c_mul}')
        print(f'Length of variation: {arc_len}')

    if plots == 'verbose':
        plot_jch(j, c_, h_)
        # plot_gamut(jab)

    return color_rgb, arc_len * c_mul


def plot_jch(j, c_, h_):
    """ Plot transition of Lightness, Chroma, Hue """
    plt.figure(figsize=(6.4, 3.6))

    ax = plt.subplot(231)
    ax.plot(j)
    ax.set_title('Lightness')
    ax.get_xaxis().set_visible(False)

    ax = plt.subplot(232)
    ax.plot(c_)
    ax.set_title('Chroma')
    ax.get_xaxis().set_visible(False)

    ax = plt.subplot(233)
    ax.plot(h_)
    ax.set_title('Hue')
    ax.get_xaxis().set_visible(False)

    ax = plt.subplot(234)
    ax.plot(np.diff(j).astype(np.float32))
    ax.set_title('ΔLightness')
    ax.get_xaxis().set_visible(False)

    ax = plt.subplot(235)
    ax.plot(np.diff(c_))
    ax.set_title('ΔChroma')
    ax.get_xaxis().set_visible(False)

    ax = plt.subplot(236)
    ax.plot(np.diff(h_[1:]))
    ax.set_title('ΔHue')
    ax.get_xaxis().set_visible(False)

    plt.show()


def plot_gamut(jab):
    """ Plot sRGB gamut for colormaps. It is broken for most of colormaps for now."""
    j = jab[..., 0]
    a = jab[..., 1]
    b = jab[..., 2]
    lightness_range = (j[0], j[-1])
    h_ = np.degrees(np.arctan2(b, a))

    # CAM16 Jab colorspace
    ac = np.linspace(-45, 45, 1000)
    bc = np.linspace(-35, 35, 1000)
    ac, bc = np.meshgrid(ac, bc)
    phi = np.degrees(np.arctan2(bc, ac))

    jc = np.searchsorted(h_, phi) / POINTS * (lightness_range[1] - lightness_range[0]) + lightness_range[0]

    xyzs = colour.CAM16UCS_to_XYZ(np.stack([jc, ac, bc], axis=-1))
    rgbs = colour.XYZ_to_sRGB(xyzs)

    # get sRGB gamut
    r = rgbs[..., 0]
    g = rgbs[..., 1]
    b = rgbs[..., 2]
    clipped = np.any([r<0, g<0, b<0, 1<r, 1<g, 1<b], axis=0)
    clipped = np.stack([clipped, clipped, clipped], axis=-1)

    rgbs[clipped] = 0.4426

    # plot gamut
    ax = plt.subplot()
    ax.plot(jab[..., 1], jab[..., 2])
    ax.axis('equal')

    ax.imshow(rgbs, extent=[ac.min(), ac.max(), bc.max(), bc.min()])
    ax.invert_yaxis()
    ax.set_title('sRGB color gamut in CAM16-UCS')
    ax.set_xlabel("a' (green → red)")
    ax.set_ylabel("b' (blue → yellow)")
    plt.show()


def plot_rgb(rgb1):
    """ Plot RGB values """
    rgbs = (rgb1*255).round().clip(0, 255).astype('uint8')

    seg_simple = 8

    fix, ax = plt.subplots()
    plt.plot(rgbs[:,0], 'r')
    plt.plot(rgbs[:,1], 'g')
    plt.plot(rgbs[:,2], 'b')
    plt.plot(np.mean(rgbs, axis=1))
    plt.title('R, G, B and mean(RGB)')

    ax.tick_params(labelbottom=False)
    ax.set_xticks(np.linspace(0, len(rgb1), seg_simple+1, endpoint=True))
    ax.set_yticks(np.arange(0, 257, 16))

    ax.grid(which='both')
    plt.show()


if __name__ == "__main__":
    PLOTS = 'verbose'

    plt.rcParams['figure.autolayout'] = True

    # hue range
    # blue to yellow -77, 111
    # AudaSpec1 = -180, 135

    # red~magenta
    # color_rgb, _ = gen_colormap([50, -70], 'circle', (3, 97), plots=PLOTS)
    # color_rgb, _ = gen_colormap([45, -65], 'square', (3, 97), plots=PLOTS)
    # color_rgb, _ = gen_colormap([75, -105], 'sin', (1, 99), plots=PLOTS)  # v
    # color_rgb, _ = gen_colormap([105, -140], 'cos', (1, 98.5), plots=PLOTS)
    # color_rgb, _ = gen_colormap([125, -180], 'cos', (2, 98.5), plots=PLOTS)

    # red blue
    # color_rgb, _ = gen_colormap([55, -175], 'square', (3, 97), plots=PLOTS)
    # color_rgb, _ = gen_colormap([95, -225], 'sin', (2, 99), plots=PLOTS)  # v
    # color_rgb, _ = gen_colormap([110, -225], 'sin', (2, 99), plots=PLOTS)
    # color_rgb, _ = gen_colormap([60, -190], 'circle', (3, 97), plots=PLOTS)
    # color_rgb, _ = gen_colormap([145, -285], 'cos', (1, 98.5), plots=PLOTS)

    # brown, purble, blue, light green, light yellow
    # color_rgb, _ = gen_colormap([80, -270], 'circle', (3, 97), plots=PLOTS)
    # color_rgb, _ = gen_colormap([75, -295], 'square', (2, 98.5), plots=PLOTS)
    # color_rgb, _ = gen_colormap([130, -320], 'sin', (2, 98.5), plots=PLOTS)  # v
    # color_rgb, _ = gen_colormap([-180, -730], 'cos', (2, 98.5), plots=PLOTS)
    # color_rgb, _ = gen_colormap([-170, -720], 'cos', (2, 98.5), plots=PLOTS)

    # rainbow
    # color_rgb, _ = gen_colormap([175, -425], 'sin', (2, 98.5), plots=PLOTS)
    # color_rgb, _ = gen_colormap([190, -495], 'sin', (2, 99), plots=PLOTS)  # v
    # color_rgb, _ = gen_colormap([-70, -945], 'cos', (2, 99), plots=PLOTS)
    # color_rgb, _ = gen_colormap([-45, -945], 'cos', (2, 99), plots=PLOTS)
    # color_rgb, _ = gen_colormap((90, -450), 'square', (3, 99), plots=PLOTS)

    # blue, pink
    # color_rgb, _ = gen_colormap((-100, 20), 'circle', (3, 97), plots=PLOTS)
    # color_rgb, _ = gen_colormap([-95, 5], 'square', (3, 97), plots=PLOTS)
    # color_rgb, _ = gen_colormap([-125, 45], 'sin', (2, 98.5), plots=PLOTS)  # v
    # color_rgb, _ = gen_colormap([-155, 75], 'cos', (2, 98.5), plots=PLOTS)

    # blue, magenta, orange, yellow
    # color_rgb, _ = gen_colormap([-139, 138], 'sin', lightness_range=(2, 99), plots=PLOTS)  # optimal
    # color_rgb, _ = gen_colormap([-185, 189], 'cos', lightness_range=(2, 99), plots=PLOTS)  # optimal
    color_rgb, _ = gen_colormap([-185, 170], 'cos', lightness_range=(2, 99), plots=PLOTS)
    # color_rgb, _ = gen_colormap([-180, 160], 'cos', lightness_range=(2, 99), plots=PLOTS)

    # color_rgb, _ = gen_colormap([10, -170], 'sin', (2, 99), plots=PLOTS)  # v

    # color_rgb, _ = gen_colormap([210, 30], 'sin', (2, 98.5), plots=PLOTS)  # v
    # color_rgb, _ = gen_colormap([-135, -335], 'cos', (2, 98.5), plots=PLOTS)

    # color_rgb, _ = gen_colormap([-75, 105], 'sin', (2, 98), plots=PLOTS)

    # color_rgb, _ = gen_colormap([70, 250], 'sin', (2, 99), plots=PLOTS)  # v
    # color_rgb, _ = gen_colormap([45, 265], 'cos', (2, 99), plots=PLOTS)

    cm_data = np.clip(color_rgb, 0, 1)

    test_cm = ListedColormap(cm_data, name='Roseus')

    from .viscm_cam16ucs import viscm
    viscm(test_cm)

    plot_rgb(color_rgb)



from importlib import import_module

from matplotlib.colors import ListedColormap

from .generator import gen_colormap, plot_rgb
from .viscm_cam16ucs import viscm


def plot_measures(name: str):
    mod = import_module(f'roseus.cmap.{name}')
    cm_data, _ = gen_colormap(*mod.gen_props, plots='verbose')

    test_cm = ListedColormap(cm_data, name=f'rs.{name}')

    viscm(test_cm)

    plot_rgb(cm_data)



from matplotlib.colors import ListedColormap

roseus: ListedColormap
roseus_r: ListedColormap
r: ListedColormap
r_r: ListedColormap
b: ListedColormap
b_r: ListedColormap
cyanus: ListedColormap
cyanus_r: ListedColormap
rbg: ListedColormap
rbg_r: ListedColormap
arcus: ListedColormap
arcus_r: ListedColormap
gr: ListedColormap
gr_r: ListedColormap
rg: ListedColormap
rg_r: ListedColormap
lavendula: ListedColormap
lavendula_r: ListedColormap



# Roseus colormap family
# https://github.com/dofuuz/roseus
# matplotlib interface
"""
matplotlib module of Roseus colormaps.

Import roseus.mpl to use with matplotlib. The colormaps will be registered to matplotlib with 'rs.' prefix.

Examples
--------
import matplotlib.pyplot as plt
import roseus.mpl as rs

plt.imshow(x, cmap=rs.arcus)
# or
plt.imshow(x, cmap='rs.arcus')
"""

import os.path as osp

import matplotlib as mpl
import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from .cmap import cmap_names


def register_colormap(name: str, rgb_data) -> tuple[Colormap, Colormap]:
    cmap = ListedColormap(rgb_data, name=f'rs.{name}')
    cmap_r = cmap.reversed()

    mpl.colormaps.register(cmap)
    mpl.colormaps.register(cmap_r)

    return cmap, cmap_r


for name in cmap_names:
    npy_path = osp.join(osp.dirname(__file__), 'cmap', f'{name}.npy')
    rgb_data = np.load(npy_path)
    cmap, cmap_r = register_colormap(name, rgb_data)

    globals()[name] = cmap
    globals()[f'{name}_r'] = cmap_r


__all__ = tuple(cmap_names) + tuple(f'{name}_r' for name in cmap_names)



# This file is part of viscm
# https://github.com/matplotlib/viscm

# Copyright (C) 2015 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2015 Stefan van der Walt <stefanv@berkeley.edu>

# Copyright (c) 2014-2016 the viscm developers
# The MIT License (MIT)

# Modified to use CAM16-UCS instead of CAM02-UCS

import os.path

import colour
import matplotlib
import matplotlib.colors
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
import numpy as np
from colorspacious import (
    cspace_convert,
    cspace_converter,
)
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas

# matplotlib.rcParams['backend'] = "QtAgg"
# Do this first before any other matplotlib imports, to force matplotlib to
# use a Qt backend
from matplotlib.colors import ListedColormap
from matplotlib.gridspec import GridSpec


GREYSCALE_CONVERSION_SPACE = "JCh"

_sRGB1_to_JCh = cspace_converter("sRGB1", GREYSCALE_CONVERSION_SPACE)
_JCh_to_sRGB1 = cspace_converter(GREYSCALE_CONVERSION_SPACE, "sRGB1")


def to_greyscale(sRGB1):
    JCh = _sRGB1_to_JCh(sRGB1)
    JCh[..., 1] = 0
    return np.clip(_JCh_to_sRGB1(JCh), 0, 1)


_deuter50_space = {"name": "sRGB1+CVD", "cvd_type": "deuteranomaly", "severity": 50}
_deuter50_to_sRGB1 = cspace_converter(_deuter50_space, "sRGB1")
_deuter100_space = {"name": "sRGB1+CVD", "cvd_type": "deuteranomaly", "severity": 100}
_deuter100_to_sRGB1 = cspace_converter(_deuter100_space, "sRGB1")
_prot50_space = {"name": "sRGB1+CVD", "cvd_type": "protanomaly", "severity": 50}
_prot50_to_sRGB1 = cspace_converter(_prot50_space, "sRGB1")
_prot100_space = {"name": "sRGB1+CVD", "cvd_type": "protanomaly", "severity": 100}
_prot100_to_sRGB1 = cspace_converter(_prot100_space, "sRGB1")


def _show_cmap(ax, rgb):
    ax.imshow(rgb[np.newaxis, ...], aspect="auto")


def _apply_rgb_mat(mat, rgb):
    return np.clip(np.einsum("...ij,...j->...i", mat, rgb), 0, 1)


# sRGB corners: a' goes from -37.4 to 45
AP_LIM = (-38, 46)
# b' goes from -46.5 to 42
BP_LIM = (-47, 43)
# J'/K goes from 0 to 100
JP_LIM = (-1, 101)


def _setup_Jpapbp_axis(ax):
    ax.set_xlabel("a' (green -> red)")
    ax.set_ylabel("b' (blue -> yellow)")
    ax.set_zlabel("J'/K (white -> black)")
    ax.set_xlim(*AP_LIM)
    ax.set_ylim(*BP_LIM)
    ax.set_zlim(*JP_LIM)


# Adapt a matplotlib colormap to a linearly transformed version -- useful for
# visualizing how colormaps look given color deficiency.
# Kinda a hack, b/c we inherit from Colormap (this is required), but then
# ignore its implementation entirely. This results in errors at runtime:
#       File "/<env>/site-packages/matplotlib/artist.py", line 1343, in format_cursor_data  # noqa: E501
#         n = self.cmap.N
#             ^^^^^^^^^^^
#     AttributeError: 'TransformedCMap' object has no attribute 'N'
class TransformedCMap(matplotlib.colors.Colormap):
    def __init__(self, transform, base_cmap):
        self.transform = transform
        self.base_cmap = base_cmap

    def __call__(self, *args, **kwargs):
        bts = kwargs.pop("bytes", False)
        fx = self.base_cmap(*args, bytes=False, **kwargs)
        tfx = self.transform(fx)
        if bts:
            return (tfx * 255).astype("uint8")
        return tfx

    def set_bad(self, *args, **kwargs):
        self.base_cmap.set_bad(*args, **kwargs)

    def set_under(self, *args, **kwargs):
        self.base_cmap.set_under(*args, **kwargs)

    def set_over(self, *args, **kwargs):
        self.base_cmap.set_over(*args, **kwargs)

    def is_gray(self):
        return False


def _vis_axes(fig):
    grid = GridSpec(
        10,
        4,
        left=0.02,
        right=0.98,
        bottom=0.02,
        width_ratios=[1] * 4,
        height_ratios=[1] * 10,
    )
    axes = {
        "cmap": grid[0, 0],
        "deltas": grid[1:4, 0],
        "cmap-greyscale": grid[0, 1],
        "lightness-deltas": grid[1:4, 1],
        "deuteranomaly": grid[4, 0],
        "deuteranopia": grid[5, 0],
        "protanomaly": grid[4, 1],
        "protanopia": grid[5, 1],
        # 'lightness': grid[4:6, 1],
        # 'colourfulness': grid[4:6, 2],
        # 'hue': grid[4:6, 3],
        "image0": grid[0:3, 2],
        "image0-cb": grid[0:3, 3],
        "image1": grid[3:6, 2],
        "image1-cb": grid[3:6, 3],
        "image2": grid[6:8, 2:],
        "image2-cb": grid[8:, 2:],
    }

    axes = {key: fig.add_subplot(value) for (key, value) in axes.items()}
    axes["gamut"] = fig.add_subplot(grid[6:, :2], projection="3d")
    return axes


def lookup_colormap_by_name(name):
    try:
        return plt.get_cmap(name)
    except ValueError:
        pass
    # Try expanding a setuptools-style entrypoint:
    #   foo.bar:baz.quux
    #   -> import foo.bar; return foo.bar.baz.quux
    if ":" in name:
        module_name, object_name = name.split(":", 1)
        object_path = object_name.split(".")
        import importlib

        cm = importlib.import_module(module_name)
        for entry in object_path:
            cm = getattr(cm, entry)
        return cm
    raise ValueError(f"Can't find colormap {name!r}")


class viscm:
    def __init__(
        self,
        cm,
        figure=None,
        uniform_space="CAM16UCS",
        name=None,
        N=256,
        N_dots=50,
        show_gamut=False,
    ):
        if isinstance(cm, str):
            cm = lookup_colormap_by_name(cm)
        if name is None:
            name = cm.name
        if figure is None:
            figure = plt.figure()
        self._sRGB1_to_uniform = lambda rgb: colour.convert(rgb, 'sRGB', uniform_space) * 100

        self.figure = figure
        self.figure.suptitle(f"Colormap evaluation: {name}", fontsize=24)

        axes = _vis_axes(self.figure)

        # ListedColormap is used for many matplotlib builtin colormaps
        # (e.g. viridis) and also what we use in the editor. It's the most
        # efficient way to work with arbitrary smooth colormaps -- pick enough
        # points that it looks smooth, and then don't waste time interpolating
        # between them. But then it creates weird issues in the analyzer if
        # our N doesn't match their N, especially when we try to compute the
        # derivative. (Specifically the derivative oscillates between 0 and a
        # large value depending on where our sample points happen to fall
        # relative to the cutoffs between the ListedColormap samples.) So if
        # this is a smooth (large N) ListedColormap, then just use its samples
        # directly:
        if isinstance(cm, ListedColormap) and cm.N >= 100:
            RGB = np.asarray(cm.colors)[:, :3]
            N = RGB.shape[0]
            x = np.linspace(0, 1, N)
        else:
            x = np.linspace(0, 1, N)
            RGB = cm(x)[:, :3]
        x_dots = np.linspace(0, 1, N_dots)
        RGB_dots = cm(x_dots)[:, :3]

        ax = axes["cmap"]
        _show_cmap(ax, RGB)
        ax.set_title("The colormap in its glory")
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        def label(ax, s):
            ax.text(
                0.95,
                0.05,
                s,
                horizontalalignment="right",
                verticalalignment="bottom",
                transform=ax.transAxes,
            )

        def title(ax, s):
            ax.text(
                0.98,
                0.98,
                s,
                horizontalalignment="right",
                verticalalignment="top",
                transform=ax.transAxes,
            )

        Jpapbp = self._sRGB1_to_uniform(RGB)

        def delta_ymax(values):
            return max(np.max(values) * 1.1, 0)

        ax = axes["deltas"]
        local_deltas = np.sqrt(np.sum((Jpapbp[:-1, :] - Jpapbp[1:, :]) ** 2, axis=-1))
        local_derivs = N * local_deltas
        ax.plot(x[1:], local_derivs)
        arclength = np.sum(local_deltas)
        rmse = np.std(local_derivs)
        title(ax, "Perceptual derivative")
        label(
            ax,
            "Length: {:0.1f}\nRMS deviation from flat: {:0.1f} ({:0.1f}%)".format(
                arclength, rmse, 100 * rmse / arclength
            ),
        )
        ax.set_ylim(-delta_ymax(-local_derivs), delta_ymax(local_derivs))
        ax.get_xaxis().set_visible(False)

        ax = axes["cmap-greyscale"]
        _show_cmap(ax, to_greyscale(RGB))
        ax.set_title("Black-and-white printed")
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        ax = axes["lightness-deltas"]
        ax.axhline(0, linestyle="--", color="grey")
        lightness_deltas = np.diff(Jpapbp[:, 0])
        lightness_derivs = N * lightness_deltas

        ax.plot(x[1:], lightness_derivs)
        title(ax, "Perceptual lightness derivative")
        lightness_arclength = np.sum(np.abs(lightness_deltas))
        lightness_rmse = np.std(lightness_derivs)
        label(
            ax,
            "Length: {:0.1f}\nRMS deviation from flat: {:0.1f} ({:0.1f}%)".format(
                lightness_arclength,
                lightness_rmse,
                100 * lightness_rmse / lightness_arclength,
            ),
        )

        ax.set_ylim(-delta_ymax(-lightness_derivs), delta_ymax(lightness_derivs))
        ax.get_xaxis().set_visible(False)

        # ax = axes['lightness']
        # ax.plot(x, ciecam02.J)
        # label(ax, "Lightness (J)")
        # ax.set_ylim(0, 105)

        # ax = axes['colourfulness']
        # ax.plot(x, ciecam02.M)
        # label(ax, "Colourfulness (M)")

        # ax = axes['hue']
        # ax.plot(x, ciecam02.h)
        # label(ax, "Hue angle (h)")
        # ax.set_ylim(0, 360)

        def anom(ax, converter, name):
            _show_cmap(ax, np.clip(converter(RGB), 0, 1))
            label(ax, name)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)

        anom(axes["deuteranomaly"], _deuter50_to_sRGB1, "Moderate deuteranomaly")
        anom(axes["deuteranopia"], _deuter100_to_sRGB1, "Complete deuteranopia")

        anom(axes["protanomaly"], _prot50_to_sRGB1, "Moderate protanomaly")
        anom(axes["protanopia"], _prot100_to_sRGB1, "Complete protanopia")

        ax = axes["gamut"]
        ax.plot(Jpapbp[:, 1], Jpapbp[:, 2], Jpapbp[:, 0])
        Jpapbp_dots = self._sRGB1_to_uniform(RGB_dots)
        ax.scatter(
            Jpapbp_dots[:, 1],
            Jpapbp_dots[:, 2],
            Jpapbp_dots[:, 0],
            c=RGB_dots[:, :],
            s=80,
        )

        # Draw a wireframe indicating the sRGB gamut
        self.gamut_patch = sRGB_gamut_patch(uniform_space)
        # That function returns a patch where each face is colored to match
        # the represented colors. For present purposes we want something
        # less... colorful.
        self.gamut_patch.set_facecolor([0.5, 0.5, 0.5, 0.1])
        self.gamut_patch.set_edgecolor([0.2, 0.2, 0.2, 0.1])
        ax.add_collection3d(self.gamut_patch)
        self.gamut_patch.set_visible(show_gamut)

        ax.view_init(elev=75, azim=-75)

        _setup_Jpapbp_axis(ax)

        images = []
        image_args = []
        example_dir = os.path.join(os.path.dirname(__file__), "viscm")

        images.append(
            np.load(os.path.join(example_dir, "st-helens_before-modified.npy")).T
        )
        image_args.append({})

        # Adapted from
        #   http://matplotlib.org/mpl_examples/images_contours_and_fields/pcolormesh_levels.py
        dx = dy = 0.05
        y, x = np.mgrid[-5 : 5 + dy : dy, -5 : 10 + dx : dx]
        z = np.sin(x) ** 10 + np.cos(10 + y * x) + np.cos(x) + 0.2 * y + 0.1 * x
        images.append(z)
        image_args.append({})

        # Peter Kovesi's colormap test image at
        #   http://peterkovesi.com/projects/colourmaps/colourmaptest.tif

        images.append(np.load(os.path.join(example_dir, "colourmaptest.npy")))

        image_args.append({})

        def _deuter_transform(RGBA):
            # clipping, alpha handling
            RGB = RGBA[..., :3]
            RGB = np.clip(_deuter50_to_sRGB1(RGB), 0, 1)
            return np.concatenate((RGB, RGBA[..., 3:]), axis=-1)

        deuter_cm = TransformedCMap(_deuter_transform, cm)

        for i, (image, args) in enumerate(zip(images, image_args)):
            ax = axes["image%i" % (i,)]
            ax.imshow(image, cmap=cm, **args)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)

            ax_cb = axes["image%i-cb" % (i,)]
            ax_cb.imshow(image, cmap=deuter_cm, **args)
            ax_cb.get_xaxis().set_visible(False)
            ax_cb.get_yaxis().set_visible(False)

        axes["image0"].set_title("Sample images")
        axes["image0-cb"].set_title("Moderate deuter.")
        self.axes = axes

    def toggle_gamut(self):
        self.gamut_patch.set_visible(not self.gamut_patch.get_visible())

    def save_figure(self, path):
        self.figure.savefig(path)


def sRGB_gamut_patch(uniform_space, resolution=20):
    step = 1.0 / resolution
    sRGB_quads = []
    sRGB_values = []
    # each entry in 'quads' is a 4x3 array where each row contains the
    # coordinates of a corner point
    for fixed in 0, 1:
        for i in range(resolution):
            for j in range(resolution):
                # R quad
                sRGB_quads.append(
                    [
                        [fixed, i * step, j * step],
                        [fixed, (i + 1) * step, j * step],
                        [fixed, (i + 1) * step, (j + 1) * step],
                        [fixed, i * step, (j + 1) * step],
                    ]
                )
                sRGB_values.append((fixed, (i + 0.5) * step, (j + 0.5) * step, 1))
                # G quad
                sRGB_quads.append(
                    [
                        [i * step, fixed, j * step],
                        [(i + 1) * step, fixed, j * step],
                        [(i + 1) * step, fixed, (j + 1) * step],
                        [i * step, fixed, (j + 1) * step],
                    ]
                )
                sRGB_values.append(((i + 0.5) * step, fixed, (j + 0.5) * step, 1))
                # B quad
                sRGB_quads.append(
                    [
                        [i * step, j * step, fixed],
                        [(i + 1) * step, j * step, fixed],
                        [(i + 1) * step, (j + 1) * step, fixed],
                        [i * step, (j + 1) * step, fixed],
                    ]
                )
                sRGB_values.append(((i + 0.5) * step, (j + 0.5) * step, fixed, 1))
    sRGB_quads = np.asarray(sRGB_quads)
    # work around colorspace transform bugginess in handling high-dim
    # arrays
    sRGB_quads_2d = sRGB_quads.reshape((-1, 3))
    Jpapbp_quads_2d = colour.convert(sRGB_quads_2d, "sRGB", uniform_space) * 100
    Jpapbp_quads = Jpapbp_quads_2d.reshape((-1, 4, 3))
    gamut_patch = mpl_toolkits.mplot3d.art3d.Poly3DCollection(
        Jpapbp_quads[:, :, [1, 2, 0]]
    )
    gamut_patch.set_facecolor(sRGB_values)
    gamut_patch.set_edgecolor(sRGB_values)
    return gamut_patch


def sRGB_gamut_Jp_slice(
    Jp, uniform_space, ap_lim=(-50, 50), bp_lim=(-50, 50), resolution=200
):
    bp_grid, ap_grid = np.mgrid[
        bp_lim[0] : bp_lim[1] : resolution * 1j, ap_lim[0] : ap_lim[1] : resolution * 1j
    ]
    Jp_grid = Jp * np.ones((resolution, resolution))
    Jpapbp = np.concatenate(
        (
            Jp_grid[:, :, np.newaxis],
            ap_grid[:, :, np.newaxis],
            bp_grid[:, :, np.newaxis],
        ),
        axis=2,
    )
    sRGB = colour.convert(Jpapbp, uniform_space / 100, "sRGB")
    sRGBA = np.concatenate((sRGB, np.ones(sRGB.shape[:2] + (1,))), axis=2)
    sRGBA[np.any((sRGB < 0) | (sRGB > 1), axis=-1)] = [0, 0, 0, 0]
    return sRGBA


def draw_pure_hue_angles(ax):
    # Pure hue angles from CIECAM-02
    for color, angle in [("r", 20.14), ("y", 90.00), ("g", 164.25), ("b", 237.53)]:
        x = np.cos(np.deg2rad(angle))
        y = np.sin(np.deg2rad(angle))
        ax.plot([0, x * 1000], [0, y * 1000], color + "--")


def draw_sRGB_gamut_Jp_slice(
    ax, Jp, uniform_space, ap_lim=(-50, 50), bp_lim=(-50, 50), **kwargs
):
    sRGB = sRGB_gamut_Jp_slice(
        Jp, uniform_space, ap_lim=ap_lim, bp_lim=bp_lim, **kwargs
    )
    im = ax.imshow(sRGB, aspect="equal", extent=ap_lim + bp_lim, origin="lower")
    draw_pure_hue_angles(ax)
    ax.set_xlim(ap_lim)
    ax.set_ylim(bp_lim)
    return im
